'use strict';

// Dependencies
var Promise = require('bluebird');
var cheerio = require('cheerio');
var moment = require('moment-timezone');
var debug = require('debug')('node-cba-netbank');

moment.tz.setDefault('Australia/Sydney');

var submittableSelector = 'input,select,textarea,keygen';
var rCRLF = /\r?\n/g;

//  Utilities
//  reference: https://github.com/cheeriojs/cheerio/blob/master/lib/api/forms.js
//  Add support for allow `disabled` and `button` input element in the serialized array.
function serializeArray(element) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : { disabled: false, button: false };

  // Resolve all form elements from either forms or collections of form elements
  return element.map(function (i, elem) {
    var $elem = cheerio(elem);
    if (elem.name === 'form') {
      return $elem.find(submittableSelector).toArray();
    }
    return $elem.filter(submittableSelector).toArray();
  }).filter(
  // Verify elements have a name (`attr.name`)
  // and are not disabled (`:disabled`) if `options.disabled === false`
  // and cannot be clicked (`[type=submit]`) if `options.button === true`
  // are used in 'x-www-form-urlencoded' ('[type=file]')
  '[name!=""]' + (options.disabled ? '' : ':not(:disabled)') + ':not(' + (options.button ? '' : ':submit, :button, ') + ':image, :reset, :file)' +
  // and are either checked/don't have a checkable state
  ':matches([checked], :not(:checkbox, :radio))'
  // Convert each of the elements to its value(s)
  ).map(function (i, elem) {
    var $elem = cheerio(elem);
    var name = $elem.attr('name');
    var value = $elem.val();

    // If there is no value set (e.g. `undefined`, `null`), then default value to empty
    if (value == null) {
      value = $elem.attr('type') === 'checkbox' ? 'on' : '';
    }

    // If we have an array of values (e.g. `<select multiple>`),
    // return an array of key/value pairs
    if (Array.isArray(value)) {
      return value.map(function (val) {
        return { name: name, value: val.replace(rCRLF, '\r\n') };
      }); //   These can occur inside of `<textarea>'s` // to guarantee consistency across platforms // We trim replace any line endings (e.g. `\r` or `\r\n` with `\r\n`)
      // Otherwise (e.g. `<input type="text">`, return only one key/value pair
    }
    return { name: name, value: value.replace(rCRLF, '\r\n') };

    // Convert our result to an array
  }).get();
}

//  parse the balance to a double number.
//  e.g. '$1,766.50 CR', '$123.00 DR'
//  Positive is CR, and negative is DR.
function parseCurrencyText(text) {
  var amount = Number(text.replace(/[^0-9.]+/g, ''));
  if (text.indexOf('DR') > -1) {
    amount = -amount;
  }
  return amount;
}

//  Parse:
// <div class="FieldElement FieldElementLabel FieldElementNoLabel">
//   <span class="CurrencySymbol CurrencyLabel PreFieldText">$</span>
//   <span title="$" class="Currency field WithPostFieldText">1,767.44</span>
//   <span class="PostFieldText">CR</span>
// </div>
//  to Number `1767.44`.
function parseCurrencyHtml(elem) {
  var e = cheerio.load('<div>' + elem + '</div>');
  var currency = parseFloat(e('.Currency').text().replace(/,/g, '').trim());
  if (e('.PostFieldText').text().trim() === 'DR') {
    return -currency;
  }
  return currency;
}

// Transaction format
// {
//  timestamp,
//  date,
//  description,
//  amount,
//  balance,
//  trancode,
//  receiptnumber
// }
function parseTransaction(json) {
  try {
    //  try parse the date from 'Date.Sort[1]' first
    var dateTag = json.Date.Sort[1];
    var t = moment.utc(dateTag, 'YYYYMMDDHHmmssSSS').tz('Australia/Sydney');
    if (!t.isValid()) {
      //  try parse the date from 'Date.Text' if previous attempt failed
      t = moment(json.Date.Text, 'DD MMM YYYY');
      //  use sort order to distinguish different transactions.
      if (dateTag && !isNaN(+dateTag)) {
        t.millisecond(+dateTag);
      }
    }

    return {
      timestamp: t.valueOf(),
      date: t.format(),
      //  replace newline with '; '
      description: json.Description.Text.replace(/\n/g, '; '),
      amount: parseCurrencyText(json.Amount.Text),
      balance: parseCurrencyText(json.Balance.Text),
      trancode: json.TranCode.Text || '',
      receiptnumber: json.ReceiptNumber.Text || ''
    };
  } catch (err) {
    //  ignore the error for the misformatted transaction, and return null.
    debug(err);
    debug('Cannot parse the given transaction: ' + JSON.stringify(json));
    return null;
  }
}

// Parsers

function parseTitle(resp) {
  return new Promise(function (resolve) {
    var $ = cheerio.load(resp.body);
    var contentType = resp.headers['content-type'];
    if (contentType && contentType.indexOf('text/html') >= 0) {
      var m = /NetBank - ([^-]+)/.exec($('title').text());
      if (m) {
        var title = m[1].trim();
        debug('parseTitle(): found title => \'' + title + '\'');
        return resolve(Object.assign({}, resp, { title: title }));
      }
    }
    debug('parseTitle(): cannot find title');
    return resolve(resp);
  });
}

function parseForm(resp) {
  return new Promise(function (resolve, reject) {
    var $ = cheerio.load(resp.body);
    var form = {};
    serializeArray($('form'), { disabled: true, button: true }).forEach(function (item) {
      form[item.name] = item.value;
    });

    if (Object.keys(form).length === 0) {
      return reject('parseForm(): Cannot find form.');
    }

    return resolve(Object.assign({}, resp, { form: Object.assign({}, resp.form, form) }));
  });
}

function parseViewState(resp) {
  return new Promise(function (resolve) {
    var form = Object.assign({}, resp.form);
    //  get new view state of asp.net
    var REGEX_VIEW_STATE = /(__VIEWSTATE|__EVENTVALIDATION|__VIEWSTATEGENERATOR)\|([^|]+)\|/g;
    var match = REGEX_VIEW_STATE.exec(resp.body);
    while (match) {
      form[match[1]] = match[2];
      match = REGEX_VIEW_STATE.exec(resp.body);
    }
    return resolve(Object.assign({}, resp, { form: Object.assign({}, resp.form, form) }));
  });
}

//  Account format:
// {
//  name,
//  link,
//  bsb,
//  account,
//  number: {bsb+account}
//  balance
// }
function parseAccountList(resp) {
  return new Promise(function (resolve, reject) {
    var $ = cheerio.load(resp.body);
    var accountRows = $('.main_group_account_row');

    if (!accountRows || accountRows.length === 0) {
      return reject('Cannot find account list.');
    }

    var accounts = [];
    accountRows.each(function (index, elem) {
      var $$ = cheerio.load(elem);
      accounts.push({
        name: $$('.NicknameField .left a').text().trim(),
        link: $$('.NicknameField .left a').attr('href'),
        bsb: $$('.BSBField .text').text().replace(/\s+/g, '').trim(),
        account: $$('.AccountNumberField .text').text().replace(/\s+/g, '').trim(),
        balance: parseCurrencyHtml($$('.AccountBalanceField')),
        available: parseCurrencyHtml($$('.AvailableFundsField'))
      });
    });

    accounts = accounts
    //  Assemble the `bsb` and `account` to construct `number`
    .map(function (acc) {
      return Object.assign({}, acc, { number: '' + acc.bsb + acc.account });
    }
    //  validate the account info
    ).filter(function (acc) {
      return acc.name && acc.link && acc.account;
    });

    debug('parseAccountList(): found ' + accounts.length + ' accounts');
    return resolve(Object.assign({}, resp, { accounts: accounts }));
  });
}

function parseHomePage(resp) {
  return parseForm(resp).then(parseTitle).then(parseAccountList);
}

function parseTransactions(resp) {
  return new Promise(function (resolve, reject) {
    //  Get transactions
    var m = /({"Transactions":(?:[^;]+))\);/.exec(resp.body);
    var transactions = void 0;
    if (m) {
      var json = JSON.parse(m[1]);
      var pendings = json.OutstandingAuthorizations.map(parseTransaction).filter(function (v) {
        return !!v;
      });
      debug('parseTransactions(): found ' + pendings.length + ' pending transactions');
      transactions = json.Transactions.map(parseTransaction).filter(function (v) {
        return !!v;
      });
      debug('parseTransactions(): found ' + transactions.length + ' transactions');
      debug('parseTransactions(): There ' + (json.More ? 'are ' : 'is NO ') + 'more transactions.');
      if (json.Limit) {
        debug('parseTransactions(): However, it reaches the limit.');
      }
      return resolve(Object.assign({}, resp, { transactions: transactions, more: json.More, limit: json.Limit, pendings: pendings }));
    }
    return reject('Cannot find transactions in the resp');
  });
}

//  Parse the account list from TransactionHistory/History.aspx page
// AccountWithKeys
// {
//  name,
//  number,
//  key
// }
function parseAccountListWithKeys(resp) {
  return new Promise(function (resolve, reject) {
    var $ = cheerio.load(resp.body);
    var accounts = [];

    $('#ctl00_ContentHeaderPlaceHolder_updatePanelAccount').find('option').each(function (i, e) {
      var texts = $(e).text().split('|').map(function (t) {
        return t.trim();
      });
      if (texts.length === 2) {
        var acc = {
          name: texts[0],
          number: texts[1].replace(/\s+/g, ''),
          key: $(e).attr('value')
        };
        if (acc.key && acc.number && acc.name) {
          accounts.push(acc);
        }
      }
    });

    if (accounts.length > 0) {
      resolve(Object.assign({}, resp, { accounts: accounts }));
    } else {
      reject('Cannot find accounts keys');
    }
  });
}

function parseTransactionPage(resp) {
  return parseForm(resp).then(parseTitle).then(parseTransactions).then(parseAccountListWithKeys);
}

module.exports = {
  serializeArray: serializeArray,
  parseTitle: parseTitle,
  parseForm: parseForm,
  parseViewState: parseViewState,
  parseAccountList: parseAccountList,
  parseHomePage: parseHomePage,
  parseCurrencyText: parseCurrencyText,
  parseCurrencyHtml: parseCurrencyHtml,
  parseTransaction: parseTransaction,
  parseTransactions: parseTransactions,
  parseAccountListWithKeys: parseAccountListWithKeys,
  parseTransactionPage: parseTransactionPage
};